import * as bunyan from 'bunyan';
import { Match, InputData, IFileService } from '@ethereum-sourcify/core';
export interface InjectorConfig {
    alchemyPID?: string;
    silent?: boolean;
    log?: bunyan;
    offline?: boolean;
    repositoryPath?: string;
    fileService?: IFileService;
}
export declare class Injector {
    private log;
    private chains;
    private alchemyPID;
    private offline;
    fileService: IFileService;
    repositoryPath: string;
    /**
     * Constructor
     * @param {InjectorConfig = {}} config
     */
    private constructor();
    /**
     * Creates an instance of Injector. Waits for chains to initialize.
     * Await this method to work with an instance that has all chains initialized.
     * @param config
     */
    static createAsync(config?: InjectorConfig): Promise<Injector>;
    /**
     * Creates an instance of Injector. Does not initialize chains.
     * @param config
     */
    static createOffline(config?: InjectorConfig): Injector;
    /**
     * Instantiates a web3 provider for all public ethereum networks via Infura/Alchemy or regular node.
     * If environment variable TESTING is set to true, localhost:8545 is also available.
     */
    private initChains;
    /**
     * Searches a set of addresses for the one whose deployedBytecode
     * matches a given bytecode string
     * @param {String[]}          addresses
     * @param {string}      deployedBytecode
     */
    private matchBytecodeToAddress;
    /**
     * Returns a string description of how closely two bytecodes match. Bytecodes
     * that match in all respects apart from their metadata hashes are 'partial'.
     * Bytecodes that don't match are `null`.
     * @param  {string} deployedBytecode
     * @param  {string} creationData
     * @param  {string} compiledRuntimeBytecode
     * @param  {string} compiledCreationBytecode
     * @param  {string} chain chainId of the chain where contract is being checked
     * @param  {string} address contract address
     * @return {Match}  match description ('perfect'|'partial'|null) and possibly constructor args (ABI-encoded) and library links
     */
    private compareBytecodes;
    private addLibraryAddresses;
    /**
     * Returns the `creationData` from the transaction that created the contract at the provided chain and address.
     * @param chain
     * @param contractAddress
     * @returns `creationData` if found, `null` otherwise
     */
    private getCreationData;
    private extractEncodedConstructorArgs;
    /**
     * Throws if addresses array contains a null value (express) or is length 0
     * @param {string[] = []} addresses param (submitted to injector)
     */
    private validateAddresses;
    /**
     * Throws if `chain` is falsy or wrong type
     * @param {string} chain param (submitted to injector)
     */
    private validateChain;
    /**
     * Used by the front-end. Accepts a set of source files and a metadata string,
     * recompiles / validates them and stores them in the repository by chain/address
     * and by swarm | ipfs hash.
     * @param  {string}            repository repository root (ex: 'repository')
     * @param  {string}            chain      chain name (ex: 'ropsten')
     * @param  {string}            address    contract address
     * @param  {string[]}          files
     * @return {Promise<object>}              address & status of successfully verified contract
     */
    inject(inputData: InputData): Promise<Match>;
    /**
     * This method exists because many different people have contributed to this code, which has led to the
     * lack of unanimous nomenclature
     * @param status
     * @returns {MatchQuality} matchQuality
     */
    private statusToMatchQuality;
    private sanitizePath;
    private getMetadataPathFromCborEncoded;
    /**
     * Stores the metadata from compilationResult to the swarm | ipfs subrepo. The exact storage path depends
     * on the swarm | ipfs address extracted from compilationResult.deployedByteode.
     *
     * @param chain used only for logging
     * @param address used only for loggin
     * @param compilationResult should contain deployedBytecode and metadata
     */
    private storeMetadata;
    /**
     * Writes the verified sources (.sol files) to the repository.
     * @param {string}              chain             chain name (ex: 'ropsten')
     * @param {string}              address           contract address
     * @param {StringMap}           sources           'rearranged' sources
     * @param {MatchQuality}        matchQuality
     */
    private storeSources;
    /**
     * Writes the constructor arguments to the repository.
     * @param matchQuality
     * @param chain
     * @param address
     * @param encodedConstructorArgs
     */
    private storeConstructorArgs;
    /**
     * Writes the map of library links (pairs of the format <placeholder:address>) to the repository.
     * @param matchQuality
     * @param chain
     * @param address
     * @param libraryMap
     */
    private storeLibraryMap;
}
