"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Injector = void 0;
const web3_1 = __importDefault(require("web3"));
const core_1 = require("@ethereum-sourcify/core");
const utils_1 = require("../utils");
// eslint-disable-next-line @typescript-eslint/no-var-requires
const multihashes = require('multihashes');
class InjectorChain {
    constructor(chain) {
        this.rpc = chain.rpc;
        this.name = chain.name;
        this.contractFetchAddress = chain.contractFetchAddress;
        this.graphQLFetchAddress = chain.graphQLFetchAddress;
        this.txRegex = chain.txRegex;
        this.archiveWeb3 = chain.archiveWeb3;
    }
}
class LoggerWrapper {
    constructor(logger) {
        this.logger = logger;
        this.logId = Math.random().toString().slice(2);
    }
    info(obj, ...params) {
        return this.logger.info(Object.assign(obj, { verificationId: this.logId }), ...params);
    }
    error(obj, ...params) {
        return this.logger.error(Object.assign(obj, { verificationId: this.logId }), ...params);
    }
}
class Injector {
    /**
     * Constructor
     * @param {InjectorConfig = {}} config
     */
    constructor(config = {}) {
        this.chains = {};
        this.alchemyPID = config.alchemyPID;
        this.offline = config.offline || false;
        this.repositoryPath = config.repositoryPath;
        this.log = config.log || core_1.Logger("Injector");
        this.fileService = config.fileService || new core_1.FileService(this.repositoryPath, this.log);
    }
    /**
     * Creates an instance of Injector. Waits for chains to initialize.
     * Await this method to work with an instance that has all chains initialized.
     * @param config
     */
    static async createAsync(config = {}) {
        const instance = new Injector(config);
        if (!instance.offline) {
            await instance.initChains();
        }
        return instance;
    }
    /**
     * Creates an instance of Injector. Does not initialize chains.
     * @param config
     */
    static createOffline(config = {}) {
        return new Injector(config);
    }
    /**
     * Instantiates a web3 provider for all public ethereum networks via Infura/Alchemy or regular node.
     * If environment variable TESTING is set to true, localhost:8545 is also available.
     */
    async initChains() {
        if (this.alchemyPID) {
            this.log.info({ loc: "[INIT_CHAINS]" }, "started checking providerPID");
            await utils_1.checkEndpoint(this.alchemyPID).catch((err) => {
                this.log.warn({ providerID: this.alchemyPID }, err.message);
            });
            this.log.info({ loc: "[INIT_CHAINS]" }, "finished checking providerPID");
        }
        const chainsData = this.alchemyPID ? core_1.getSupportedChains() : core_1.getFullnodeChains();
        for (const chain of chainsData) {
            this.chains[chain.chainId] = new InjectorChain(chain);
            if (this.alchemyPID) {
                const web3 = chain.rpc[0];
                this.chains[chain.chainId].web3 = new web3_1.default(web3);
            }
            else {
                const web3 = chain.fullnode.dappnode;
                this.chains[chain.chainId].web3 = new web3_1.default(web3);
                await utils_1.checkEndpoint(web3).catch(() => {
                    this.log.warn({ endpoint: web3 }, `Invalid endpoint for chain ${chain.name}`);
                });
            }
        }
        const numberOfChains = Object.keys(this.chains).length;
        this.log.info({ loc: "[INIT_CHAINS]", numberOfChains }, "Finished loading chains");
    }
    /**
     * Searches a set of addresses for the one whose deployedBytecode
     * matches a given bytecode string
     * @param {String[]}          addresses
     * @param {string}      deployedBytecode
     */
    async matchBytecodeToAddress(chain, addresses = [], recompiled) {
        let match = { address: null, status: null };
        const chainName = this.chains[chain].name || "The chain";
        for (let address of addresses) {
            address = web3_1.default.utils.toChecksumAddress(address);
            let deployedBytecode = null;
            try {
                this.log.info({
                    loc: '[MATCH]',
                    chain: chain,
                    address: address
                }, `Retrieving contract bytecode address`);
                deployedBytecode = await utils_1.getBytecode(this.chains[chain].web3, address);
            }
            catch (e) { /* ignore */ }
            try {
                match = await this.compareBytecodes(deployedBytecode, null, recompiled, chain, address);
            }
            catch (err) {
                if (addresses.length === 1) {
                    match.message = "There were problems during contract verification. Please try again in a minute.";
                }
            }
            if (match.status) {
                break;
            }
            else if (addresses.length === 1 && !match.message) {
                if (!deployedBytecode) {
                    match.message = `${chainName} is temporarily unavailable.`;
                }
                else if (deployedBytecode === "0x") {
                    match.message = `${chainName} does not have a contract deployed at ${address}.`;
                }
                else {
                    match.message = "The deployed and recompiled bytecode don't match.";
                }
            }
        }
        return match;
    }
    /**
     * Returns a string description of how closely two bytecodes match. Bytecodes
     * that match in all respects apart from their metadata hashes are 'partial'.
     * Bytecodes that don't match are `null`.
     * @param  {string} deployedBytecode
     * @param  {string} creationData
     * @param  {string} compiledRuntimeBytecode
     * @param  {string} compiledCreationBytecode
     * @param  {string} chain chainId of the chain where contract is being checked
     * @param  {string} address contract address
     * @return {Match}  match description ('perfect'|'partial'|null) and possibly constructor args (ABI-encoded) and library links
     */
    async compareBytecodes(deployedBytecode, creationData, recompiled, chain, address) {
        const match = {
            address,
            status: null,
            encodedConstructorArgs: undefined,
            libraryMap: undefined
        };
        if (deployedBytecode && deployedBytecode.length > 2) {
            const { replaced, libraryMap } = this.addLibraryAddresses(recompiled.deployedBytecode, deployedBytecode);
            recompiled.deployedBytecode = replaced;
            match.libraryMap = libraryMap;
            if (deployedBytecode === recompiled.deployedBytecode) {
                match.status = "perfect";
                return match;
            }
            const trimmedDeployedBytecode = utils_1.getBytecodeWithoutMetadata(deployedBytecode);
            const trimmedCompiledRuntimeBytecode = utils_1.getBytecodeWithoutMetadata(recompiled.deployedBytecode);
            if (trimmedDeployedBytecode === trimmedCompiledRuntimeBytecode) {
                match.status = "partial";
                return match;
            }
            if (trimmedDeployedBytecode.length === trimmedCompiledRuntimeBytecode.length) {
                creationData = creationData || await this.getCreationData(chain, address);
                const { replaced, libraryMap } = this.addLibraryAddresses(recompiled.creationBytecode, creationData);
                recompiled.creationBytecode = replaced;
                match.libraryMap = libraryMap;
                if (creationData) {
                    if (creationData.startsWith(recompiled.creationBytecode)) {
                        // The reason why this uses `startsWith` instead of `===` is that
                        // creationData may contain constructor arguments at the end part.
                        const encodedConstructorArgs = this.extractEncodedConstructorArgs(creationData, recompiled.creationBytecode);
                        match.status = "perfect";
                        match.encodedConstructorArgs = encodedConstructorArgs;
                        return match;
                    }
                    const trimmedCompiledCreationBytecode = utils_1.getBytecodeWithoutMetadata(recompiled.creationBytecode);
                    if (creationData.startsWith(trimmedCompiledCreationBytecode)) {
                        match.status = "partial";
                        return match;
                    }
                }
            }
        }
        return match;
    }
    addLibraryAddresses(template, real) {
        const PLACEHOLDER_START = "__$";
        const PLACEHOLDER_LENGTH = 40;
        const libraryMap = {};
        let index = template.indexOf(PLACEHOLDER_START);
        for (; index !== -1; index = template.indexOf(PLACEHOLDER_START)) {
            const placeholder = template.slice(index, index + PLACEHOLDER_LENGTH);
            const address = real.slice(index, index + PLACEHOLDER_LENGTH);
            libraryMap[placeholder] = address;
            const regexCompatiblePlaceholder = placeholder.replace("__$", "__\\$").replace("$__", "\\$__");
            const regex = RegExp(regexCompatiblePlaceholder, "g");
            template = template.replace(regex, address);
        }
        return {
            replaced: template,
            libraryMap
        };
    }
    /**
     * Returns the `creationData` from the transaction that created the contract at the provided chain and address.
     * @param chain
     * @param contractAddress
     * @returns `creationData` if found, `null` otherwise
     */
    async getCreationData(chain, contractAddress) {
        const loc = "[GET_CREATION_DATA]";
        let txFetchAddress = this.chains[chain].contractFetchAddress;
        const txRegex = this.chains[chain].txRegex;
        if (txFetchAddress && txRegex) { // fetch from a block explorer and extract by regex
            txFetchAddress = txFetchAddress.replace("${ADDRESS}", contractAddress);
            const web3 = this.chains[chain].web3;
            this.log.info({ loc, chain, contractAddress, fetchAddress: txFetchAddress }, "Scraping block explorer");
            try {
                return await utils_1.getCreationDataByScraping(txFetchAddress, txRegex, web3);
            }
            catch (err) {
                this.log.error({ loc, chain, contractAddress, err: err.message }, "Scraping failed!");
            }
        }
        const graphQLFetchAddress = this.chains[chain].graphQLFetchAddress;
        if (graphQLFetchAddress) { // fetch from graphql node
            const web3 = this.chains[chain].web3;
            try {
                return await utils_1.getCreationDataFromGraphQL(graphQLFetchAddress, contractAddress, web3);
            }
            catch (err) {
                this.log.error({ loc, chain, contractAddress, err: err.message });
            }
        }
        const archiveWeb3 = this.chains[chain].archiveWeb3;
        if (archiveWeb3) { // fetch by binary search on chain history
            this.log.info({ loc, chain, contractAddress }, "Fetching archive data");
            try {
                return await utils_1.getCreationDataFromArchive(contractAddress, archiveWeb3);
            }
            catch (err) {
                this.log.error({ loc, chain, contractAddress, err: err.message }, "Archive search failed!");
            }
        }
        const err = "Cannot fetch creation data";
        this.log.error({ loc, chain, contractAddress, err });
        throw new Error(err);
    }
    extractEncodedConstructorArgs(creationData, compiledCreationBytecode) {
        const startIndex = creationData.indexOf(compiledCreationBytecode);
        return "0x" + creationData.slice(startIndex + compiledCreationBytecode.length);
    }
    /**
     * Throws if addresses array contains a null value (express) or is length 0
     * @param {string[] = []} addresses param (submitted to injector)
     */
    validateAddresses(addresses = []) {
        const err = new Error("Missing address for submitted sources/metadata");
        if (!addresses.length) {
            throw err;
        }
        for (const address of addresses) {
            if (address == null)
                throw err;
        }
    }
    /**
     * Throws if `chain` is falsy or wrong type
     * @param {string} chain param (submitted to injector)
     */
    validateChain(chain) {
        if (!chain || typeof chain !== 'string') {
            throw new Error("Missing chain name for submitted sources/metadata");
        }
    }
    /**
     * Used by the front-end. Accepts a set of source files and a metadata string,
     * recompiles / validates them and stores them in the repository by chain/address
     * and by swarm | ipfs hash.
     * @param  {string}            repository repository root (ex: 'repository')
     * @param  {string}            chain      chain name (ex: 'ropsten')
     * @param  {string}            address    contract address
     * @param  {string[]}          files
     * @return {Promise<object>}              address & status of successfully verified contract
     */
    async inject(inputData) {
        const { chain, addresses, contract } = inputData;
        this.validateAddresses(addresses);
        this.validateChain(chain);
        let match;
        const wrappedLogger = new LoggerWrapper(this.log);
        if (!core_1.CheckedContract.isValid(contract)) {
            await core_1.CheckedContract.fetchMissing(contract, wrappedLogger);
        }
        const compilationResult = await utils_1.recompile(contract.metadata, contract.solidity, wrappedLogger);
        // When injector is called by monitor, the bytecode has already been
        // obtained for address and we only need to compare w/ compilation result.
        if (inputData.bytecode) {
            if (addresses.length !== 1) {
                const err = "Injector cannot work with multiple addresses if bytecode is provided";
                this.log.error({ loc: "[INJECTOR]", addresses, err });
                throw new Error(err);
            }
            const address = web3_1.default.utils.toChecksumAddress(addresses[0]);
            match = await this.compareBytecodes(inputData.bytecode, inputData.creationData, compilationResult, chain, address);
            // For other cases, we need to retrieve the code for specified address
            // from the chain.
        }
        else {
            match = await this.matchBytecodeToAddress(chain, addresses, compilationResult);
        }
        // Since the bytecode matches, we can be sure that we got the right
        // metadata file (up to json formatting) and exactly the right sources.
        // Now we can store the re-compiled and correctly formatted metadata file
        // and the sources.
        if (match.address && match.status) {
            const metadataPath = this.getMetadataPathFromCborEncoded(compilationResult, match.address, chain);
            if (metadataPath) {
                this.fileService.save(metadataPath, compilationResult.metadata);
                this.fileService.deletePartial(chain, match.address);
            }
            else {
                match.status = "partial";
            }
            const matchQuality = this.statusToMatchQuality(match.status);
            this.storeSources(matchQuality, chain, match.address, contract.solidity);
            this.storeMetadata(matchQuality, chain, match.address, compilationResult);
            if (match.encodedConstructorArgs && match.encodedConstructorArgs.length) {
                this.storeConstructorArgs(matchQuality, chain, match.address, match.encodedConstructorArgs);
            }
            if (match.libraryMap && Object.keys(match.libraryMap).length) {
                this.storeLibraryMap(matchQuality, chain, match.address, match.libraryMap);
            }
        }
        else {
            const message = match.message || "Could not match the deployed and recompiled bytecode.";
            const err = new Error(`Contract name: ${contract.name}. ${message}`);
            this.log.error({
                loc: '[INJECT]', chain, addresses, err
            });
            throw new Error(err.message);
        }
        return match;
    }
    /**
     * This method exists because many different people have contributed to this code, which has led to the
     * lack of unanimous nomenclature
     * @param status
     * @returns {MatchQuality} matchQuality
     */
    statusToMatchQuality(status) {
        if (status === "perfect")
            return "full";
        if (status === "partial")
            return status;
    }
    sanitizePath(originalPath) {
        return originalPath
            .replace(/[^a-z0-9_./-]/gim, "_")
            .replace(/(^|\/)[.]+($|\/)/, '_');
    }
    getMetadataPathFromCborEncoded(compilationResult, address, chain) {
        const bytes = web3_1.default.utils.hexToBytes(compilationResult.deployedBytecode);
        const cborData = core_1.cborDecode(bytes);
        if (cborData['bzzr0']) {
            return `/swarm/bzzr0/${web3_1.default.utils.bytesToHex(cborData['bzzr0']).slice(2)}`;
        }
        else if (cborData['bzzr1']) {
            return `/swarm/bzzr1/${web3_1.default.utils.bytesToHex(cborData['bzzr1']).slice(2)}`;
        }
        else if (cborData['ipfs']) {
            return `/ipfs/${multihashes.toB58String(cborData['ipfs'])}`;
        }
        this.log.error({
            loc: '[INJECTOR:GET_METADATA_PATH]',
            address,
            chain,
            err: "No metadata hash in cbor encoded data."
        });
        return null;
    }
    /**
     * Stores the metadata from compilationResult to the swarm | ipfs subrepo. The exact storage path depends
     * on the swarm | ipfs address extracted from compilationResult.deployedByteode.
     *
     * @param chain used only for logging
     * @param address used only for loggin
     * @param compilationResult should contain deployedBytecode and metadata
     */
    storeMetadata(matchQuality, chain, address, compilationResult) {
        this.fileService.save({
            matchQuality,
            chain,
            address,
            fileName: "metadata.json"
        }, compilationResult.metadata);
    }
    /**
     * Writes the verified sources (.sol files) to the repository.
     * @param {string}              chain             chain name (ex: 'ropsten')
     * @param {string}              address           contract address
     * @param {StringMap}           sources           'rearranged' sources
     * @param {MatchQuality}        matchQuality
     */
    storeSources(matchQuality, chain, address, sources) {
        for (const sourcePath in sources) {
            this.fileService.save({
                matchQuality,
                chain,
                address,
                source: true,
                fileName: this.sanitizePath(sourcePath)
            }, sources[sourcePath]);
        }
    }
    /**
     * Writes the constructor arguments to the repository.
     * @param matchQuality
     * @param chain
     * @param address
     * @param encodedConstructorArgs
     */
    storeConstructorArgs(matchQuality, chain, address, encodedConstructorArgs) {
        this.fileService.save({
            matchQuality,
            chain,
            address,
            source: false,
            fileName: "constructor-args.txt"
        }, encodedConstructorArgs);
    }
    /**
     * Writes the map of library links (pairs of the format <placeholder:address>) to the repository.
     * @param matchQuality
     * @param chain
     * @param address
     * @param libraryMap
     */
    storeLibraryMap(matchQuality, chain, address, libraryMap) {
        const indentationSpaces = 2;
        this.fileService.save({
            matchQuality,
            chain,
            address,
            source: false,
            fileName: "library-map.json"
        }, JSON.stringify(libraryMap, null, indentationSpaces));
    }
}
exports.Injector = Injector;
//# sourceMappingURL=Injector.js.map