"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getCreationDataFromGraphQL = exports.getCreationDataByScraping = exports.getCreationDataFromArchive = exports.getSolcJs = exports.getBytecodeWithoutMetadata = exports.recompile = exports.getBytecode = exports.checkEndpoint = void 0;
const web3_1 = __importDefault(require("web3"));
const node_fetch_1 = __importDefault(require("node-fetch"));
const core_1 = require("@ethereum-sourcify/core");
const path_1 = __importDefault(require("path"));
const fs_1 = __importDefault(require("fs"));
// eslint-disable-next-line @typescript-eslint/no-var-requires
const solc = require('solc');
const child_process_1 = require("child_process");
const http_status_codes_1 = require("http-status-codes");
const ethers_1 = require("ethers");
const GITHUB_SOLC_REPO = "https://github.com/ethereum/solc-bin/raw/gh-pages/linux-amd64/";
/**
 * Checks if provided endpoint is online
 * @param {string} provider personal project ID from infura.io or real Ethereum node endpoint
 */
async function checkEndpoint(provider) {
    if (provider.includes("http")) {
        const web3 = new web3_1.default(provider);
        await web3.eth.getNodeInfo().catch(() => {
            throw new Error("Check your node");
        });
    }
    else if (provider) {
        await node_fetch_1.default(`https://eth-mainnet.alchemyapi.io/v2/${provider}`, {
            method: "post",
            headers: {
                'Accept': 'application/json',
                'Content-Type': 'application/json'
            },
            body: JSON.stringify({ "jsonrpc": "2.0", "id": 1, "method": "eth_blockNumber", "params": [] })
        })
            .then((response) => {
            if (response.status == http_status_codes_1.StatusCodes.UNAUTHORIZED) {
                throw new Error("Check your Infura ID");
            }
        }).catch(() => {
            throw new Error("Check your Infura ID");
        });
    }
    else {
        throw new Error("No provider set");
    }
}
exports.checkEndpoint = checkEndpoint;
/**
 * Wraps eth_getCode
 * @param {Web3}   web3    connected web3 instance
 * @param {string} address contract
 */
async function getBytecode(web3, address) {
    address = web3.utils.toChecksumAddress(address);
    return await web3.eth.getCode(address);
}
exports.getBytecode = getBytecode;
const RECOMPILATION_ERR_MSG = "Recompilation error (probably caused by invalid metadata)";
/**
 * Compiles sources using version and settings specified in metadata
 * @param  {any}                          metadata
 * @param  {string[]}                     sources  solidity files
 * @return {Promise<RecompilationResult>}
 */
async function recompile(metadata, sources, log) {
    const { input, fileName, contractName } = core_1.reformatMetadata(metadata, sources, log);
    const loc = "[RECOMPILE]";
    const version = metadata.compiler.version;
    log.info({ loc, fileName, contractName, version }, 'Recompiling');
    const compiled = await useCompiler(version, input, log);
    const output = JSON.parse(compiled);
    if (!output.contracts || !output.contracts[fileName] || !output.contracts[fileName][contractName]) {
        const errors = output.errors.filter((e) => e.severity === "error").map((e) => e.message);
        log.error({ loc, fileName, contractName, version, errors });
        throw new Error(RECOMPILATION_ERR_MSG);
    }
    const contract = output.contracts[fileName][contractName];
    return {
        creationBytecode: `0x${contract.evm.bytecode.object}`,
        deployedBytecode: `0x${contract.evm.deployedBytecode.object}`,
        metadata: contract.metadata.trim()
    };
}
exports.recompile = recompile;
/**
 * Searches for a solc: first for a local executable version, then from GitHub
 * and then using the getSolcJs function.
 * Once the compiler is retrieved, it is used, and the stringified solc output is returned.
 *
 * @param version the version of solc to be used for compilation
 * @param input a JSON object of the standard-json format compatible with solc
 * @param log the logger
 * @returns stringified solc output
 */
async function useCompiler(version, input, log) {
    const inputStringified = JSON.stringify(input);
    const solcPath = await getSolcExecutable(version, log);
    let compiled = null;
    if (solcPath) {
        const logObject = { loc: "[RECOMPILE]", version, solcPath };
        log.info(logObject, "Compiling with external executable");
        const shellOutputBuffer = child_process_1.spawnSync(solcPath, ["--standard-json"], { input: inputStringified });
        if (!shellOutputBuffer.stdout) {
            log.error(logObject, shellOutputBuffer.error || RECOMPILATION_ERR_MSG);
            throw new Error(RECOMPILATION_ERR_MSG);
        }
        compiled = shellOutputBuffer.stdout.toString();
    }
    else {
        const soljson = await getSolcJs(version, log);
        compiled = soljson.compile(inputStringified);
    }
    return compiled;
}
function validateSolcPath(solcPath, log) {
    const spawned = child_process_1.spawnSync(solcPath, ["--version"]);
    if (spawned.status === 0) {
        return true;
    }
    const error = spawned.error ? spawned.error.message : "Unknown error";
    log.error({ loc: "[VALIDATE_SOLC_PATH]", solcPath, error });
    return false;
}
async function getSolcExecutable(version, log) {
    const fileName = `solc-linux-amd64-v${version}`;
    const tmpSolcRepo = process.env.SOLC_REPO_TMP || path_1.default.join("/tmp", "solc-repo");
    const repoPaths = [tmpSolcRepo, process.env.SOLC_REPO || "solc-repo"];
    for (const repoPath of repoPaths) {
        const solcPath = path_1.default.join(repoPath, fileName);
        if (fs_1.default.existsSync(solcPath) && validateSolcPath(solcPath, log)) {
            return solcPath;
        }
    }
    const tmpSolcPath = path_1.default.join(tmpSolcRepo, fileName);
    const success = await fetchSolcFromGitHub(tmpSolcPath, version, fileName, log);
    return success ? tmpSolcPath : null;
}
async function fetchSolcFromGitHub(solcPath, version, fileName, log) {
    const githubSolcURI = GITHUB_SOLC_REPO + encodeURIComponent(fileName);
    const logObject = { loc: "[RECOMPILE]", version, githubSolcURI };
    log.info(logObject, "Fetching executable solc from GitHub");
    const res = await node_fetch_1.default(githubSolcURI);
    if (res.status === http_status_codes_1.StatusCodes.OK) {
        log.info(logObject, "Successfully fetched executable solc from GitHub");
        fs_1.default.mkdirSync(path_1.default.dirname(solcPath), { recursive: true });
        const buffer = await res.buffer();
        try {
            fs_1.default.unlinkSync(solcPath);
        }
        catch (_e) {
            undefined;
        }
        fs_1.default.writeFileSync(solcPath, buffer, { mode: 0o755 });
        if (validateSolcPath(solcPath, log)) {
            return true;
        }
    }
    else {
        log.error(logObject, "Failed fetching executable solc from GitHub");
    }
    return false;
}
/**
 * Removes post-fixed metadata from a bytecode string
 * (for partial bytecode match comparisons )
 * @param  {string} bytecode
 * @return {string}          bytecode minus metadata
 */
function getBytecodeWithoutMetadata(bytecode) {
    // Last 4 chars of bytecode specify byte size of metadata component,
    const metadataSize = parseInt(bytecode.slice(-4), 16) * 2 + 4;
    return bytecode.slice(0, bytecode.length - metadataSize);
}
exports.getBytecodeWithoutMetadata = getBytecodeWithoutMetadata;
/**
 * Fetches the requested version of the Solidity compiler (soljson).
 * First attempts to search locally; if that fails, falls back to downloading it.
 *
 * @param version the solc version to retrieve: the expected format is
 *
 * "[v]<major>.<minor>.<patch>+commit.<hash>"
 *
 * e.g.: "0.6.6+commit.6c089d02"
 *
 * defaults to "latest"
 *
 * @param log a logger to track the course of events
 *
 * @returns the requested solc instance
 */
function getSolcJs(version = "latest", log) {
    // /^\d+\.\d+\.\d+\+commit\.[a-f0-9]{8}$/
    version = version.trim();
    if (version !== "latest" && !version.startsWith("v")) {
        version = "v" + version;
    }
    const soljsonRepo = process.env.SOLJSON_REPO || "soljson-repo";
    const soljsonPath = path_1.default.resolve(soljsonRepo, `soljson-${version}.js`);
    log.info({ loc: "[GET_SOLC]", "target": soljsonPath }, "Searching for js solc locally");
    if (fs_1.default.existsSync(soljsonPath)) {
        log.info({ loc: "[GET_SOLC]" }, "Found js solc locally");
        return new Promise(resolve => {
            // eslint-disable-next-line @typescript-eslint/no-var-requires
            const soljson = solc.setupMethods(require(soljsonPath));
            resolve(soljson);
        });
    }
    log.info({ loc: "[GET_SOLC]", version }, "Searching for js solc remotely");
    return new Promise((resolve, reject) => {
        solc.loadRemoteVersion(version, (error, soljson) => {
            if (error) {
                log.error({ loc: "[GET_SOLC]", version }, "Could not find solc remotely");
                reject(error);
            }
            else {
                log.info({ loc: "[GET_SOLC]", version }, "Found solc remotely");
                resolve(soljson);
            }
        });
    });
}
exports.getSolcJs = getSolcJs;
async function getCreationBlockNumber(contractAddress, web3) {
    let highestBlock = await web3.eth.getBlockNumber();
    let lowestBlock = 0;
    let contractCode = await web3.eth.getCode(contractAddress, highestBlock);
    if (contractCode == "0x") {
        throw new Error(`Contract ${contractAddress} does not exist!`);
    }
    while (lowestBlock <= highestBlock) {
        const searchBlock = Math.floor((lowestBlock + highestBlock) / 2);
        contractCode = await web3.eth.getCode(contractAddress, searchBlock);
        if (contractCode != "0x") {
            highestBlock = searchBlock;
        }
        else if (contractCode == "0x") {
            lowestBlock = searchBlock;
        }
        if (highestBlock == lowestBlock + 1) {
            return highestBlock;
        }
    }
}
/**
 * Modified and optimized version of https://www.shawntabrizi.com/ethereum-find-contract-creator/
 *
 * @param contractAddress the hexadecimal address of the contract
 * @param web3 initialized web3 object for chain requests
 * @returns a Promise of the creation data
 */
async function getCreationDataFromArchive(contractAddress, web3) {
    const creationBlockNumber = await getCreationBlockNumber(contractAddress, web3);
    const creationBlock = await web3.eth.getBlock(creationBlockNumber, true);
    const transactions = creationBlock.transactions;
    for (const i in transactions) {
        const transaction = transactions[i];
        const calculatedContractAddress = ethers_1.ethers.utils.getContractAddress(transaction);
        if (calculatedContractAddress === contractAddress) {
            return transaction.input;
        }
    }
    throw new Error(`Creation data of contract ${contractAddress} could not be located!`);
}
exports.getCreationDataFromArchive = getCreationDataFromArchive;
/**
 * Returns the data used for contract creation in the transaction found by the provided regex on the provided page.
 *
 * @param fetchAddress the URL from which to fetch the page to be scrapd
 * @param txRegex regex whose first group matches the transaction hash on the page
 * @param web3 initialized web3 object for chain requests
 * @returns a promise of the creation data
 */
async function getCreationDataByScraping(fetchAddress, txRegex, web3) {
    const res = await node_fetch_1.default(fetchAddress);
    const buffer = await res.buffer();
    const page = buffer.toString();
    if (res.status === http_status_codes_1.StatusCodes.OK) {
        const matched = page.match(txRegex);
        if (matched && matched[1]) {
            const txHash = matched[1];
            const tx = await web3.eth.getTransaction(txHash);
            return tx.input;
        }
    }
    throw new Error(`Creation data could not be scraped from ${fetchAddress}`);
}
exports.getCreationDataByScraping = getCreationDataByScraping;
async function getCreationDataFromGraphQL(fetchAddress, contractAddress, web3) {
    const body = JSON.stringify({ query: `
        query AccountCreationTx {
        allAccounts(first:1, filter: {
            address:{
            equalTo:"${contractAddress.toLowerCase()}"
            },
            creationTx: {
            isNull: false
            }
        }) {
            nodes {
            creationTx
            }
        }
        }`
    });
    const rawResponse = await node_fetch_1.default(fetchAddress, {
        method: 'POST',
        headers: {
            'Accept': 'application/json',
            'Content-Type': 'application/json'
        },
        body
    });
    try {
        const resp = await rawResponse.json();
        const txHash = resp.data.allAccounts.nodes[0].creationTx;
        const tx = await web3.eth.getTransaction(txHash);
        return tx.input;
    }
    catch (err) {
        throw new Error(`Creation data could not be fetched from ${fetchAddress}, reason: ${err.message}`);
    }
}
exports.getCreationDataFromGraphQL = getCreationDataFromGraphQL;
//# sourceMappingURL=utils.js.map